-- CardGenerator 
--
-- Description: Generates cards from four simulated decks.
--              Reshuffles at the beginning of each hand.
--
-- comments:
--   lfsb info: http://homepage.mac.com/afj/taplist.html
--

-- hds interface_start
--  CardGenerator 
-- 
--  Description: Generates cards from four simulated decks.
--               Reshuffles at the beginning of each hand.
-- 
--  comments:
--    lfsb info: http://homepage.mac.com/afj/taplist.html
-- 
-- 
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

ENTITY card_generator IS
   PORT( 
      in_clock         : IN     std_logic;
      in_reset         : IN     std_logic;                      -- reset to new game
      in_issue         : IN     std_logic;                      -- request a new card
      in_bist          : IN     std_logic;                      -- test mode (standard cards)
      main_reset       : IN     std_logic;
      -- New signals to be implemented(2006)-------------
      cgIssueHP        : IN     STD_LOGIC;
      issuedHP         : OUT    STD_LOGIC;   --currntly issued card is Hidden card for the house player(by 2006)
      ---------------------------------------------
      shuffle          : IN     std_logic;                      --shuffle (change random_seed)
      dealttotal_out   : IN     std_logic_vector (7 DOWNTO 0);
      in_num_twos      : IN     std_logic_vector (4 DOWNTO 0);
      in_num_threes    : IN     std_logic_vector (4 DOWNTO 0);
      in_num_fours     : IN     std_logic_vector (4 DOWNTO 0);
      in_num_fives     : IN     std_logic_vector (4 DOWNTO 0);
      in_num_sixes     : IN     std_logic_vector (4 DOWNTO 0);
      in_num_sevens    : IN     std_logic_vector (4 DOWNTO 0);
      in_num_eights    : IN     std_logic_vector (4 DOWNTO 0);
      in_num_nines     : IN     std_logic_vector (4 DOWNTO 0);
      in_num_tens      : IN     std_logic_vector (4 DOWNTO 0);
      in_num_jacks     : IN     std_logic_vector (4 DOWNTO 0);
      in_num_kings     : IN     std_logic_vector (4 DOWNTO 0);
      in_num_queens    : IN     std_logic_vector (4 DOWNTO 0);
      in_num_aces      : IN     std_logic_vector (4 DOWNTO 0);
      out_card         : OUT    std_logic_vector (3 DOWNTO 0);  -- card output on issue
      out_house        : OUT    std_logic_vector (3 DOWNTO 0);  -- dealer card on reset
      out_issued       : OUT    std_logic;                      -- signals out_card is ready (pulse high)
      over_reset       : OUT    std_logic;
      ccard            : OUT    std_logic_vector (3 DOWNTO 0);
      house_card_dealt : OUT    std_logic
   );

-- Declarations

END card_generator ;
-- hds interface_end

architecture card_generator_arch of card_generator is 
signal random_seed : std_logic_vector(15 downto 0); --extract to a signal
signal tmp_card_check : std_logic_vector(3 downto 0) ;
signal regenerate: std_logic := '0';   -- this signal is used for regenerate a card when necessary; for observation only(added 2007)

begin
    
rand_seed: process(in_clock) -- this process is simply for random_seed generation(2007)
begin
if (main_reset = '1') then
   random_seed <="0001110100000000";
else
     if(in_clock'event and in_clock = '1') then
    -- random_seed generated by counter. random reset captures a counter state
    	   if( random_seed = "1111111111111111" ) then	--make sure lfsr doesn't get stuck w/ all 0's
        random_seed <= "0000000000000001";
        else
        random_seed <= random_seed + 1;    
        end if;
    end if;
end if;
end process;
 
  process(in_clock, in_issue, in_reset)
    variable startup_flag : std_logic ; -- flag for startup
    variable dealt_house_flag : std_logic ; -- flag for having dealt the first card after shuffle
    --variable random_seed : std_logic_vector(15 downto 0); -- seed for lfsb
    variable intern :std_logic_vector(15 downto 0);
    
    variable lfsr : std_logic_vector(15 downto 0) ; -- linear feedback shift register
    variable lfsr_next_bit : std_logic ; -- linear feedback shift register next bit
    variable card : std_logic_vector(3 downto 0) ; -- card bits that are verified and output
    variable tmp_card : std_logic_vector(3 downto 0) ;
    variable sequence : natural range 0 to 10; -- cyclic sequence used for correcting bad cards
    variable sum_of_played : std_logic_vector(4 downto 0) ; -- sum of played cards (from stats)
    
     -- memory for output
    variable var_out_card : std_logic_vector(3 downto 0) ;
    variable var_out_house : std_logic_vector(3 downto 0) ;
    variable var_out_issued : std_logic ;
    variable over_reset_in :  std_logic ;
    variable house_card_dealt_in : std_logic;
    variable startup_bist :std_logic ;
    variable regenerate_var: std_logic := '0';  
     --regenerate is the flag to determine whether the generated card is eligible to put out.
         begin

    --begin clocked statement for free-running clock and reset
    if ( main_reset ='1') then
      startup_flag := '0';
      dealt_house_flag := '0';
      intern :="0001110100000000";
      lfsr := (others=>'0');
      lfsr_next_bit := '0';
      card := (others=>'0');
      tmp_card	:= (others=>'0'); 
      sum_of_played := (others=>'0');
      var_out_card := (others=>'0');
      var_out_house := (others=>'0');
      var_out_issued  := '0';
    	 over_reset_in  := '0';
    	 house_card_dealt_in :='0';
    	 startup_bist  :='0';
	   issuedHP <= '0';
	   regenerate <= '0';
      sequence := 1;
      regenerate_var := '0';
  else
        
   -- if(in_clock'event and in_clock = '1') then
         -- random_seed generated by counter. random reset captures a counter state
    --	random_seed := random_seed + 1;    	 
   -- end if;
    
    house_card_dealt_in :='0';
    over_reset_in :='0';
    house_card_dealt <= house_card_dealt_in;
    
    	if ( in_reset'event and in_reset = '0') then	--reset signal starts new shuffle

		--get random seed at boot-up or at end of bist
	    	if( in_bist = '0' AND startup_flag = '0' ) then 		
	    		lfsr := random_seed;
  		   startup_flag := '1';   --shows the random_seed is read to the lfsr(2007) 		
  		   startup_bist := '0';
	    	elsif( in_bist = '1' AND startup_bist ='0') then -- in bist, use non-random lfsb state
  		   lfsr := "1110010100011100";
	    	 intern :=lfsr;
	    	 startup_bist := '1';   --shows the bist initial lfsr value is read to lfsr(2007)
      	 	startup_flag := '0';	    	
	    	else
	    	null;-- if neither boot nor bist, let lfsr do its thing
	    	end if; 
	    	regenerate_var := '0';	    	   	
	    	dealt_house_flag := '0';
	  end if; -- end in_reset rising edge
	        
	        

     -- card generation is clocked on the in_issue signal
     -------------------------------------------Start Generating------------------------
    if( (in_issue'event and in_issue = '1') OR dealt_house_flag = '0' OR regenerate_var = '1' )then
                 
        over_reset_in :='0';
    	
    	-- update cyclic sequence
    	if( sequence = 10 ) then
    		sequence := 1;
    	else
    		sequence := sequence + 1;
    	end if;

	-- build potential card value from lfsr (0 to 12 range)
  	for i in 1 to 4 loop
  		card(i-1) := lfsr(15);
 		lfsr_next_bit := lfsr(15) XOR lfsr(14) XOR lfsr(12) XOR lfsr(3); 
  		lfsr(15 downto 1) := lfsr(14 downto 0);
  		lfsr(0) := lfsr_next_bit;
  		
 	   if( in_bist = '1' ) then	
	    	if (lfsr ="0000000000000000") then
  		     if( intern = "0000000000000000" ) then  		           
  		           lfsr :=(others=>'1');
			  else
             		  lfsr:=intern;
	    	   end if;	
   	       intern := intern+7561;
  		   end if;  		
       end if;
  		
   end loop;
 	
 	card := card + 1;-- adjust to (1 to 13) scale     	
      	-- check for non-card values and generate a card from the non-card (0 to 12 range)
      	if( card > 12 ) then
      	regenerate_var := '1';
      	--	card := (14 - card) + sequence; -- (14-card) = [0,2] and sequence=[0,10]
      	else
      	regenerate_var := '0';
      	end if;
				
	-- check for validity of card (i.e. decks haven't been exhausted) and correct
      	--  loop through stats until finding a legal card (does nothing if already legal)
      	--  this is unnecessary for the house card since it is the first card
      	if( dealt_house_flag = '0' and regenerate_var = '0' ) then
      		var_out_house := card;	--deal house card
      		dealt_house_flag := '1';
      		var_out_issued := '0';
      		house_card_dealt_in :='1';
   	    house_card_dealt <= house_card_dealt_in;
      		ccard <= var_out_house;   --ccard includes all card generated
      	else
	      	tmp_card := card;
		
		--for i in 1 to 13 loop

	    		case tmp_card is 
				when "0001" => sum_of_played := in_num_aces;
				when "0010" => sum_of_played := in_num_twos;
				when "0011" => sum_of_played := in_num_threes;
				when "0100" => sum_of_played := in_num_fours;
				when "0101" => sum_of_played := in_num_fives;
				when "0110" => sum_of_played := in_num_sixes;
				when "0111" => sum_of_played := in_num_sevens;
				when "1000" => sum_of_played := in_num_eights;
				when "1001" => sum_of_played := in_num_nines;
				when "1010" => sum_of_played := in_num_tens;
				when "1011" => sum_of_played := in_num_jacks;
				when "1100" => sum_of_played := in_num_queens;
				when "1101" => sum_of_played := in_num_kings;
				when others => sum_of_played := in_num_fives;
			end case;
		
		if( sum_of_played >= "10000" ) then
		    regenerate_var := '1';
		 --  tmp_card_check <=tmp_card;
			  --if( tmp_card = "1101" ) then
				 --	tmp_card :="0001";
			 -- else
				  -- tmp_card := tmp_card+'1';
--			  end if;
		--	  if (dealttotal_out ="11010000") then
			 --      over_reset_in :='1';   --?necessary?
			  --end if;	       
		--	tmp_card_check <=tmp_card;
		else
			card := tmp_card;
		end if;

	--	end loop;
		if(regenerate_var = '0') then
    		var_out_card := card; -- deal player card
	   var_out_issued := '1';
  	   ccard <= var_out_card;
  	   end if;
 end if;-- end dealt_house_flag
end if; -- end in_issue clock
---------------------------------------Output------------------	
  	-- we issue a 'issuedHP' to inform the controller that the currntly issued card is Hidden card for the house player
	-- in response to the controller signal cgissueHP
if (regenerate_var = '0') then
	if (cgIssueHP = '1' ) then
		  issuedHP <='1';
	else issuedHP <='0';		
	end if;

	-- out_issued should go low after in_issue goes low
	if( in_issue'event and in_issue = '0' ) then
		var_out_issued := '0';
	end if;
	--assign all outputs every cycle
    out_card <= var_out_card;
    out_house <= var_out_house;
    out_issued <= var_out_issued;

  end if; -- end regenerate_var
  over_reset <= ( over_reset_in or shuffle );
  regenerate <= regenerate_var;
 end if;  --end main_reset
end process;
  
 
  
  
end card_generator_arch;
									 

